# frozen_string_literal: true

module Hermes
  # Score class
  class Score
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::ActivableConcern

    ORIENTATION_BOOLS = %w[
      single_parent
      priority_public
      mutation
    ].freeze

    attr_accessor :detailed_score_intermediate

    # applicant weighting
    @@under_30_year_points = 70
    @@rejected_cal_points = 40
    @@declined_cal_points = -40
    
    # request
    @@valid_motivation_points = 70
    @@valid_motivations = ['19', '28']
    @@seniority_per_month_points = 1
    
    # applicant entity weighting
    @@validate_and_active_points = 240

    # priority_points
    @@priority_points_thresholds = [30, 70, 145]

    field :department, type: String
    field :sne_date, type: DateTime
    field :profile_score, type: Integer
    field :last_eligible_position, type: Integer
    field :epci, type: Integer
    field :profile_epci_score, type: Integer
    # score en comptant la position candidature
    field :position_points, type: Integer
    field :has_applicated, type: Boolean, default: false
    field :eligible_position, type: Integer, default: 0
    field :applicated_position, type: Integer
    field :effort_rate, type: Float
    field :priority_public, type: Boolean, default: false
    field :single_parent, type: Boolean, default: false
    field :mutation, type: Boolean, default: false
    field :reserved, type: Boolean, default: false
    field :applicant_department, type: String
    field :applicant_postal_code, type: String
    field :entity_id, type: BSON::ObjectId
    field :housing_application_id, type: BSON::ObjectId
    field :apl_amount, type: Float

    common_indexes *%i[
      is_active
      department
      sne_date
      profile_score
      last_eligible_position
      epci
      profile_epci_score
      position_points
      has_applicated
      eligible_position
      applicated_position
      effort_rate
      priority_public
      single_parent
      mutation
      reserved
      applicant_department
      applicant_postal_code
      entity_id
      housing_application_id
      apl_amount
    ]

    belongs_to(
      :applicant,
      class_name: 'Hermes::Applicant',
      index: true
    )
    belongs_to(
      :housing_offer,
      class_name: 'Hermes::HousingOffer',
      index: true
    )

    embeds_one :detailed_score, class_name: 'Hermes::DetailedScore'

    validates_presence_of :applicant_id
    validates_presence_of :housing_offer_id

    scope :inactives, -> { where(is_active: false) }
    scope :applications, -> { where(has_applicated: true) }

    before_save :before_actions
    after_update :after_actions
    after_create :after_create_actions
    after_save :after_save_actions
    after_destroy :after_destroy_actions

    def located_eligible_position
      total_by_department  = Hermes::Score.where(housing_offer_id: housing_offer_id, applicant_department: applicant.department)
      total_by_postal_code = Hermes::Score.where(housing_offer_id: housing_offer_id, applicant_postal_code: applicant.postal_code)
      {
        position_by_department: eligible_position_for(total_by_department),
        total_by_department: total_by_department.pluck(:id).count,
        position_by_postal_code: eligible_position_for(total_by_postal_code),
        total_by_postal_code: total_by_postal_code.pluck(:id).count
      }
    end

    def eligible_position_for(location_query)
      if location_query.count.positive?
        location_query.where(
          id: { :$ne => id },
          eligible_position: { :$lt => eligible_position }
        ).pluck(:id).count + 1
      else
        0
      end
    end

    def active_request
      applicant&.first_active_request
    end

    def entity
      applicant&.entity
    end

    def monthly_revenues
      active_request&.monthly_revenues || 0.0
    end

    def rent_with_charges
      housing_offer&.rent_with_charges
    end

    def coapplicants
      return unless (request = active_request)

      request.coapplicants_raw || []
    end

    def applicant_is_under_30
      !!applicant.birth_date && applicant.birth_date + 30.years >= DateTime.current.utc 
    end

    def entity_points
      return 0 if !entity || (active_request && Hermes::HousingRequest.remove_entity_stock_status.include?(active_request.entity_status))

      points = 0
      if active_request.priority_points? && !
        points += active_request.priority_points_indexes.inject(0) do |sum, index|
          sum + index
        end
      end

      points + entity_validation_points
    end

    def entity_validation_points
      # wrike 461636692
      points = 0
      points += @@validate_and_active_points if applicant&.first_active_request.is_validated?
      points
    end

    def calculate_effort_rate
      return 0 unless monthly_revenues.to_i.positive? && rent_with_charges.present?
      rent = rent_with_charges.to_f
      rent -= apl_amount.to_f if !HERMES_CONFIG['effort_rate_without_apl'] && apl_amount
      rent / monthly_revenues
    end

    def calculate_effort_rate_points
      if effort_rate >= 0.26 && effort_rate <= 0.33
        40
      else
        #  (effort_rate.to_f * 100 * 40 / 26).floor
        0
      end
    end

    def already_rejected_in_cal
      active_request.rejected_in_cal_nb&.send('>', 0)
    end

    def already_declined_cal
      active_request.rejected_after_cal_nb&.send('>', 0)
    end

    def calculate_history_points
      score = 0
      score += @@rejected_cal_points if already_rejected_in_cal
      score += @@declined_cal_points if already_declined_cal
      score
    end

    def calculate_under_30_year_points
      score = 0
      score += @@under_30_year_points if applicant_is_under_30
      score
    end

    def calculate_seniority_points
      score = 0
      request_created_at = active_request.sne_created_at || active_request.created_at
      now = Time.now
      months = (now.year * 12 + now.month) - (request_created_at.year * 12 + request_created_at.month)
      score += months * @@seniority_per_month_points
      score
    end

    def calculate_motivation_points
      motivations = (active_request) ? [active_request.motivation1, active_request.motivation2, active_request.motivation3] : []
      (motivations & @@valid_motivations).empty? ? 0 : @@valid_motivation_points
    end 

    def calculate_score
      @detailed_score_intermediate = {}

      return unless active_request && housing_offer

      detailed_score_intermediate[:under_30_year_points] = calculate_under_30_year_points
      detailed_score_intermediate[:history_points] = calculate_history_points
      detailed_score_intermediate[:entity_points] = entity_points
      detailed_score_intermediate[:effort_rate_points] = calculate_effort_rate_points
      detailed_score_intermediate[:motivation_points] = calculate_motivation_points
      detailed_score_intermediate[:seniority_points] = calculate_seniority_points

      score = calculate_under_30_year_points
      score += calculate_history_points
      score += entity_points
      score += calculate_motivation_points
      score += calculate_seniority_points

      score + calculate_effort_rate_points
    end

    def build_age(applicant) 
      if applicant.birth_date == nil
        return nil 
      else 
        age = Date.today.year - applicant.birth_date.year
        age -= 1 if Date.today < applicant.birth_date + age.years
        return age
      end 
    end 

    def build_month
      date = "#{Date.today.year}-#{Date.today.month + 1}"
      return date
    end 

    def build_year
      date = Date.today.year - 2
      return date
    end 


    def build_individus
      link = Hermes::ApplicantLink.where(applicant_ids: self.applicant.id).to_a.first
      if link == nil 
        applicant = {}

        applicant["#{self.applicant.first_name}"] = 
                                                    {
                                                      "age": {
                                                      "#{build_month}": build_age(self.applicant)
                                                      }, 
                                                      "revenu_assimile_salaire": {
                                                        "#{build_year}": self.applicant.amount_year_minus_two
                                                      },
                                                      "enfant_a_charge":{
                                                        "#{build_year}": self.applicant.valid_children_nb
                                                      }
                                                    }
                    
        return applicant
      else 
        applicants = {}
        link.applicant_ids.each do |applicant_id|
          applicant = Hermes::Applicant.find(applicant_id)
          new_applicant = {
                              "age": {
                              "#{build_month}": build_age(applicant)
                              }, 
                              "revenu_assimile_salaire": {
                                "#{build_year}": applicant.amount_year_minus_two
                              },
                              "enfant_a_charge":{
                                "#{build_year}": 4
                              }
                            }
                          
          applicants["#{applicant.first_name}"] = new_applicant
        end 
        return applicants
      end 
    end 

    def build_parents
      list = []
      build_individus.each do |individu|
        list << individu.first
      end
      return list 
    end 

    def build_coloc
      link = Hermes::ApplicantLink.where(applicant_ids: self.applicant.id).to_a.first
      if !link.nil?
        if link.type == "L"
          return true 
        else 
          return false
        end 
      else 
        return false 
      end 
    end 

    def build_couple 
      link = Hermes::ApplicantLink.where(applicant_ids: self.applicant.id).to_a.first
      if !link.nil?
        if link.type == "M" || link.type == "P" || link.type == "C"
          return true 
        else 
          return false 
        end 
      else 
        return false 
      end 
    end 

    def build_apl_body
      body = {
        "individus": build_individus,
        "menages": {
          "menage_1": {
            "personne_de_reference": [
              build_parents.first
            ],
            "loyer": {
              "#{build_month}": self.housing_offer.rent_amount
            },
            "zone_apl": {
              "#{build_month}": build_zone_apl
            },
            "statut_occupation_logement":{
              "#{build_month}": "locataire_vide"
            }, 
            "coloc":{
              "#{build_month}": build_coloc

            }
          }
        },
        "familles": {
          "famille_1": {
            "parents": build_parents,
          "aide_logement_montant": {
          "#{build_month}": nil
          }, 
          "aide_logement_base_ressources": {
            "#{build_month}": nil
          },
          "en_couple": {
            "#{build_month}": build_couple
          }
        }
        },
        "foyers_fiscaux": {
          "foyer_fiscal_1": {
            "declarants": build_parents.first
          }
        }
      }
      return body 
    end 

    def build_zone_apl
      city = self.housing_offer.city 
      match = false 
      zone_2 = [
            "marseille",
            "lyon",
            "toulouse",
            "nice",
            "nantes",
            "strasbourg",
            "montpellier",
            "bordeaux",
            "rennes",
            "le havre",
            "reims",
            "lille",
            "saint-etienne",
            "toulon",
            "angers",
            "brest",
            "grenoble",
            "dijon",
            "le mans",
            "clermont-ferrand"]

      if city.nil?
        zone_apl = "zone_2"
      else 
        if city.include?("paris")
          match = true 
          zone_apl = "zone_1"
        end

        unless match == true 
          zone_2.each do |zone_2_city| 
            unless match == true 
              if city.include?(zone_2_city)
                match = true 
                zone_apl = "zone_2"
              end 
            end 
          end 
        end 
        unless match == true 
          zone_apl = "zone_3"
        end 
      end 
      return zone_apl
    end 

    def calculate_apl_amount
      unless self.applicant == nil
        uri = URI('http://149.202.173.114:5000/calculate')
        http = Net::HTTP.new(uri.host, uri.port)
        req = Net::HTTP::Post.new(uri.path, 'Content-Type' => 'application/json')
        req.body = build_apl_body.to_json
        res = http.request(req)
        body = JSON.parse(res.body)
        apl_amount = body['familles']['famille_1']['aide_logement_montant']["#{build_month}"]
        self.apl_amount = apl_amount
        self.save
      end 
    end 

    def applicated?
      !!active_request && !!Hermes::HousingApplication.find_by(is_active: true, housing_request: active_request, housing_offer: housing_offer)
    end

    def update_has_applicated
      self.has_applicated = applicated?
    end

    # set sne_date to the oldest of nuds creation_date matching the offer's department
    def update_sne_date
      self.sne_date = active_request&.nuds
                                    &.where(department: housing_offer.department)
                                    &.order_by(creation_date: 1)
                                    &.first
                                    &.creation_date
    end

    def update_profile_score
      self.profile_score = calculate_score
    end

    def update_department
      self.department = housing_offer.department
    end

    def update_effort_rate
      self.effort_rate = calculate_effort_rate
    end

    def update_profile_epci_score
      self.profile_epci_score = (profile_score || 0) + (epci || 0)
    end

    def inherit_data
      if active_request
        ORIENTATION_BOOLS.each do |orientation_name|
          send(
            "#{orientation_name}=",
            active_request.send(orientation_name)
          )
        end
        self.reserved = active_request.entity_validated
      end
      self.entity_id = entity&.id if reserved
      self.applicant_postal_code = applicant&.postal_code
      self.applicant_department = applicant&.department
    end

    def before_actions
      inherit_data
      update_has_applicated
      update_department
      update_effort_rate
      update_profile_score
      update_profile_epci_score
      update_sne_date
      update_housing_application_id
    end

    def process
      before_actions
      save if self.changed?
    end 

    def update_housing_application_id
      return unless active_request
      application = Hermes::HousingApplication.find_by(
        is_active: true, 
        housing_offer_id: housing_offer_id, 
        housing_request_id: active_request.id 
      )

      self.housing_application_id = application ? application.id : nil 
    end

    def orientation_status
      orientation = ORIENTATION_BOOLS.each_with_object({}) do |orientation_name, h|
        h[orientation_name.to_sym] = if send(orientation_name)
                                       1
                                     else
                                       -1
                                     end
      end
      orientation[:entities] = { entity_id.to_s => 1 } if reserved
      orientation
    end

    def inverted_orientation_status
      orientation_status.inverted_values
    end

    def unnegated_inverted_orientation_status
      orientation_status.unnegate.inverted_values
    end

    def orientation_changes
      changes = ORIENTATION_BOOLS.each_with_object({}) do |orientation_name, h|
        if send("#{orientation_name}_changed?")
          h[orientation_name.to_sym] = if send(orientation_name)
                                         -1
                                       else
                                         1
                                       end
        end
      end
      if reserved_changed?
        changes[:entities] = { entity_id.to_s => reserved ? 1 : -1 }
      elsif reserved && entity_id_changed?
        changes[:entities] = {
          entity_id_was.to_s => -1,
          entity_id.to_s => 1
        }
      end
      changes
    end

    def orientation_changed?
      orientation_changes.present?
    end

    def handle_orientation_movements(for_deactivation = false)
      return unless housing_offer_id

      changes = if for_deactivation
                  unnegated_inverted_orientation_status
                elsif created_at == updated_at || new_record?
                  orientation_status.unnegate
                else
                  orientation_changes
                end
      HandleOfferEligibleNbJob.perform_later(housing_offer_id.to_s, changes)
    end

    def update_detailed_score
      detailed_score = build_detailed_score unless detailed_score
      data = {}
      %i[under_30_year_points history_points entity_points effort_rate_points motivation_points seniority_points].each do |attribute|
        data[attribute] = detailed_score_intermediate[attribute] if detailed_score_intermediate[attribute].present?
      end

      detailed_score.update(data) unless data.empty?
    end

    def after_actions
      if is_active_changed? && !is_active
        handle_orientation_movements(true)
      elsif orientation_changed?
        handle_orientation_movements
      end
      if profile_epci_score_changed? || 
        eligible_position_changed? || 
        applicated_position_changed? || 
        effort_rate_changed? || 
        is_active_changed? ||
        housing_application_id_changed?
        update_application_score 
      end
    end

    def after_destroy_actions
      handle_orientation_movements(true)
    end

    def after_create_actions
      after_actions
      Hermes::LogJob.perform_later(
        code: 'SC_C',
        action: "new score_created with applicant #{applicant_id.to_s} and housing_offer #{housing_offer_id.to_s}",
        score_id: id.to_s
      )
    end

    def after_save_actions
      update_detailed_score
      update_rank if has_applicated_changed? || profile_epci_score_changed? || is_active_changed? || sne_date_changed?
      delete_score_position_cache
      true
    end

    def update_rank
      UpdateRankJob.perform_later(housing_offer_id.to_s)
    end

    def update_application_score
      return unless housing_application_id && application = Hermes::HousingApplication.find(housing_application_id)
      application.update_application_score
      application.save
    end

    def notify_applicants
      # Temporary_disabled_devops_7390
      # if has_applicated_changed? && has_applicated
      #   if applicated_position && applicated_position_was && (applicated_position > applicated_position_was)
      #     offer.notify(
      #       key: NOTIFICATION_KEYS[:housing_offer][:position_decreased],
      #       target_ids: [applicant.id],
      #       datas: {
      #         score_id: id.to_s
      #       },
      #       channels: ['email']
      #     )
      #   end
      # end
    end

    def detailed_cache_keys
      [
        "detailed_score:no_applicant_id:#{housing_offer.id}",
        "detailed_score:#{applicant.id}:#{housing_offer.id}"
      ]
    end

    def delete_score_position_cache
      detailed_cache_keys.each { |key| Rails.cache.delete(key)  }
    end
  end
end
