require 'rails_helper'
require 'vcr_config'

module Hermes
  require [Hermes::Engine.root, %w[lib hermes tools specs]].flatten.join('/')
  require_relative '../../swagger_helper'

  ACCEPTABLE_PARAMS = %w[
    right
    type
    title
    group_id
    entity_id
    lessor_id
    user_id
  ]

  MUTATE_PARAMS = {
    right: 'String',
    type: 'String',
    entity_id: 'String',
    lessor_id: 'String',
    user_id: 'String',
  }

  VALID_OPERATORS = %w[
    $in
    $nin
    $ne
    $gt
    $gte
    $lt
    $lte
    $regex
  ]

  describe Api::Dmo::UserRightsController, :type => :api do
    routes {Hermes::Engine.routes}
    include Devise::Test::ControllerHelpers
    include JwtAuthHelper
      
    let!(:entity1) {create(:entity, siret: 'Super000000001', title: 'entity1')}
    let!(:entity2) {create(:entity, siret: 'Super000000002', title: 'entity2')}
    let!(:entity3) {create(:entity, siret: 'Super000000003', title: 'entity3')}
    let!(:entity4) {create(:entity, siret: 'Super000000003', title: 'entity4')}
    let!(:bailleur) { create(:user, auth_roles: ['bailleur'], first_name: 'Bailleur') }
    let!(:user_right1) { create(:user_right, user_id: bailleur.id, entity: entity1.id, type: 'entity', right: 'administrator') }
    let!(:user_right2) { create(:user_right, user_id: bailleur.id, entity: entity2.id, type: 'entity', right: 'administrator') }
      
    let!(:user1) { create(:user, auth_roles: ['Demandeur'])}
  

    describe 'POST /user_rights' do
      describe 'Create as bailleur' do
        before :each do
          authenticated_header(request, bailleur)
          post 'create', {
            data: {
              attributes: { 
                right: 'administrator',
                type: 'entity',
                entity_id: entity1.id.to_s,
                user_id: user1.id.to_s,
              }
            }
          }
        end
        it 'returns status code 201' do
          expect(response).to have_http_status(201)
          expect(json['data']['attributes']['title']).to eq('entity1')
          expect(json['data']['attributes']['user_first_name']).to eq(user1.first_name)
          expect(Hermes::Entity.find_by(id: entity1.id).writable_by?(user1.reload)).to eq(true)
        end
      end

      describe 'Error when no rights' do
        before :each do
          authenticated_header(request, user1)
          post 'create', {
            data: {
              attributes: { 
                right: 'administrator',
                type: 'entity',
                entity_id: entity1.id.to_s,
                user_id: user1.id.to_s,
              }
            }
          }
        end
        it 'returns status code 401' do
          expect(response).to have_http_status(401)
        end
      end
    end

    describe 'PUT /user_rights/:id' do 
      describe 'as bailleur' do
        before :each do
          authenticated_header(request, bailleur)
          patch 'update', { id: user_right1.id,
            data: {
              attributes: { 
                right: 'reader',
              }
            }
          }
        end
        it 'returns status code 200' do
          expect(response).to have_http_status(200)
          expect(json['data']['attributes']['title']).to eq('entity1')
          expect(json['data']['attributes']['user_first_name']).to eq(bailleur.first_name)
          expect(Hermes::Entity.find_by(id: entity1.id).writable_by?(bailleur.reload)).to eq(false)
          expect(Hermes::Entity.find_by(id: entity1.id).readable_by?(bailleur.reload)).to eq(true)
        end
      end

      describe 'Cannot change user' do
        before :each do
          authenticated_header(request, bailleur)
          patch 'update', { id: user_right1.id,
            data: {
              attributes: { 
                user_id: user1,
              }
            }
          }
        end
        it 'returns status code 422' do
          expect(response).to have_http_status(422)
        end
      end

      describe 'Error when no rights' do
        before :each do
          authenticated_header(request, user1)
          patch 'update', { id: user_right1.id,
            data: {
              attributes: { 
                right: 'reader',
              }
            }
          }
        end
        it 'returns status code 401' do
          expect(response).to have_http_status(401)
        end
      end
    end 

    describe 'DELETE /user_rights/:id' do
      describe 'Delete as bailleur' do
        before :each do
          authenticated_header(request, bailleur)
          delete 'destroy', {
            id: user_right1.id.to_s
          }
        end
        it 'returns status code 204' do
          expect(response).to have_http_status(204)
          expect(Hermes::Entity.find_by(id: entity1.id).writable_by?(bailleur.reload)).to eq(false)
        end
      end

      describe 'Error when no rights' do
        before :each do
          authenticated_header(request, user1)
          delete 'destroy', {
            id: user_right1.id.to_s
          }
        end
        it 'returns status code 401' do
          expect(response).to have_http_status(401)
        end
      end
    end

    describe 'GET /user_rights/:id' do
      describe 'Show as bailleur' do
        before :each do
          authenticated_header(request, bailleur)
          get 'show', {
            id: user_right1.id.to_s
          }
        end
        it 'returns status code 200' do
          datas = json['data']
          expect(response).to have_http_status(200)
          expect(datas['attributes']['title']).to eq entity1.title
        end
      end
    end

    describe 'GET /user_rights' do
      describe 'Index as bailleur' do
        before :each do
          authenticated_header(request, bailleur)
          get 'index'
        end
        it 'returns status code 200' do
          datas = json['data']
          expect(response).to have_http_status(200)
          expect(datas.length).to eq 2
          expect(datas[0]['attributes']['title']).to eq entity1.title
        end
      end

      describe 'Index as bailleur filter by entity_id' do
        before :each do
          authenticated_header(request, bailleur)
          get 'index', {entity_id: entity1.id.to_s}
        end
        it 'returns status code 200' do
          datas = json['data']
          expect(response).to have_http_status(200)
          expect(datas.length).to eq 1
        end
      end
    end

    describe 'POST /user_rights/search' do 
      it 'works' do 
        authenticated_header(request, bailleur)
        post 'search', {entity_id: entity1.id.to_s}
        datas = json['data']
        expect(response).to have_http_status(200)
        expect(datas.length).to eq 1
      end 
    end 

    describe 'DELETE /user_rights/bulk' do 
      before do 
        authenticated_header(request, bailleur)
      end 
      it 'works' do 
        delete 'delete_bulk', {user_right_ids: [user_right1.id.to_s]}
        expect(response).to have_http_status(204)
        expect(Hermes::UserRight.all.count).to eq(1)
      end 

      it 'when not found' do 
        delete 'delete_bulk', {user_right_ids: [user_right1.id.to_s, "dsqdqs"]}
        expect(response).to have_http_status(404)
      end 

      it 'when not right' do 
        ur = Hermes::UserRight.create(
          user_id: user1.id,
          entity_id: entity4.id,
          type: 'entity',
          right: 'administrator'
        )
        delete 'delete_bulk', {user_right_ids: [user_right1.id.to_s, ur.id.to_s]}
        expect(response).to have_http_status(401)
      end 

    end 

    describe 'POST /user_rights/bulk' do 
      let!(:params) do 
        {
          user_rights: [
            {user_id: user1.id.to_s, entity_id: entity1.id.to_s, type: 'entity', right: 'administrator'},
            {user_id: user1.id.to_s, entity_id: entity2.id.to_s, type: 'entity', right: 'administrator'},
          ]
        }
      end 

      before do 
        authenticated_header(request, bailleur)
      end 

      it 'works' do 
        post 'post_bulk', params
        datas = json['data']
        expect(response).to have_http_status(200)
        expect(datas.length).to eq 2
        expect(user1.reload.group_ids.length).to eq(2)
        expect(Hermes::UserRight.where(user: user1).count).to eq(2)
      end
      
      it 'when missing params' do 
        params[:user_rights][0].delete(:user_id)
        post 'post_bulk', params
        expect(response).to have_http_status(422)
      end 

      it 'when no right' do 
        params[:user_rights][0][:entity_id] = entity3.id.to_s
        post 'post_bulk', params
        expect(response).to have_http_status(401)
      end 

    end 
  end
end
